@@ -1,6 +1,6 @@
 /***
  * ASM: a very small and fast Java bytecode manipulation framework
- * Copyright (c) 2000-2005 INRIA, France Telecom
+ * Copyright (c) 2000-2007 INRIA, France Telecom
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,155 +35,207 @@ package org.objectweb.asm;
  * file format. It can be used alone, to generate a Java class "from scratch",
  * or with one or more {@link ClassReader ClassReader} and adapter class visitor
  * to generate a modified class from one or more existing Java classes.
- * 
+ *
  * @author Eric Bruneton
  */
 public class ClassWriter implements ClassVisitor {
 
+    /**
+     * Flag to automatically compute the maximum stack size and the maximum
+     * number of local variables of methods. If this flag is set, then the
+     * arguments of the {@link MethodVisitor#visitMaxs visitMaxs} method of the
+     * {@link MethodVisitor} returned by the {@link #visitMethod visitMethod}
+     * method will be ignored, and computed automatically from the signature and
+     * the bytecode of each method.
+     *
+     * @see #ClassWriter(int)
+     */
+    public static final int COMPUTE_MAXS = 1;
+
+    /**
+     * Flag to automatically compute the stack map frames of methods from
+     * scratch. If this flag is set, then the calls to the
+     * {@link MethodVisitor#visitFrame} method are ignored, and the stack map
+     * frames are recomputed from the methods bytecode. The arguments of the
+     * {@link MethodVisitor#visitMaxs visitMaxs} method are also ignored and
+     * recomputed from the bytecode. In other words, computeFrames implies
+     * computeMaxs.
+     *
+     * @see #ClassWriter(int)
+     */
+    public static final int COMPUTE_FRAMES = 2;
+
     /**
      * The type of instructions without any argument.
      */
-    final static int NOARG_INSN = 0;
+    static final int NOARG_INSN = 0;
 
     /**
      * The type of instructions with an signed byte argument.
      */
-    final static int SBYTE_INSN = 1;
+    static final int SBYTE_INSN = 1;
 
     /**
      * The type of instructions with an signed short argument.
      */
-    final static int SHORT_INSN = 2;
+    static final int SHORT_INSN = 2;
 
     /**
      * The type of instructions with a local variable index argument.
      */
-    final static int VAR_INSN = 3;
+    static final int VAR_INSN = 3;
 
     /**
      * The type of instructions with an implicit local variable index argument.
      */
-    final static int IMPLVAR_INSN = 4;
+    static final int IMPLVAR_INSN = 4;
 
     /**
      * The type of instructions with a type descriptor argument.
      */
-    final static int TYPE_INSN = 5;
+    static final int TYPE_INSN = 5;
 
     /**
      * The type of field and method invocations instructions.
      */
-    final static int FIELDORMETH_INSN = 6;
+    static final int FIELDORMETH_INSN = 6;
 
     /**
-     * The type of the INVOKEINTERFACE instruction.
+     * The type of the INVOKEINTERFACE/INVOKEDYNAMIC instruction.
      */
-    final static int ITFMETH_INSN = 7;
+    static final int ITFDYNMETH_INSN = 7;
 
     /**
      * The type of instructions with a 2 bytes bytecode offset label.
      */
-    final static int LABEL_INSN = 8;
+    static final int LABEL_INSN = 8;
 
     /**
      * The type of instructions with a 4 bytes bytecode offset label.
      */
-    final static int LABELW_INSN = 9;
+    static final int LABELW_INSN = 9;
 
     /**
      * The type of the LDC instruction.
      */
-    final static int LDC_INSN = 10;
+    static final int LDC_INSN = 10;
 
     /**
      * The type of the LDC_W and LDC2_W instructions.
      */
-    final static int LDCW_INSN = 11;
+    static final int LDCW_INSN = 11;
 
     /**
      * The type of the IINC instruction.
      */
-    final static int IINC_INSN = 12;
+    static final int IINC_INSN = 12;
 
     /**
      * The type of the TABLESWITCH instruction.
      */
-    final static int TABL_INSN = 13;
+    static final int TABL_INSN = 13;
 
     /**
      * The type of the LOOKUPSWITCH instruction.
      */
-    final static int LOOK_INSN = 14;
+    static final int LOOK_INSN = 14;
 
     /**
      * The type of the MULTIANEWARRAY instruction.
      */
-    final static int MANA_INSN = 15;
+    static final int MANA_INSN = 15;
 
     /**
      * The type of the WIDE instruction.
      */
-    final static int WIDE_INSN = 16;
+    static final int WIDE_INSN = 16;
 
     /**
      * The instruction types of all JVM opcodes.
      */
-    static byte[] TYPE;
+    static final byte[] TYPE;
 
     /**
      * The type of CONSTANT_Class constant pool items.
      */
-    final static int CLASS = 7;
+    static final int CLASS = 7;
 
     /**
      * The type of CONSTANT_Fieldref constant pool items.
      */
-    final static int FIELD = 9;
+    static final int FIELD = 9;
 
     /**
      * The type of CONSTANT_Methodref constant pool items.
      */
-    final static int METH = 10;
+    static final int METH = 10;
 
     /**
      * The type of CONSTANT_InterfaceMethodref constant pool items.
      */
-    final static int IMETH = 11;
+    static final int IMETH = 11;
 
     /**
      * The type of CONSTANT_String constant pool items.
      */
-    final static int STR = 8;
+    static final int STR = 8;
 
     /**
      * The type of CONSTANT_Integer constant pool items.
      */
-    final static int INT = 3;
+    static final int INT = 3;
 
     /**
      * The type of CONSTANT_Float constant pool items.
      */
-    final static int FLOAT = 4;
+    static final int FLOAT = 4;
 
     /**
      * The type of CONSTANT_Long constant pool items.
      */
-    final static int LONG = 5;
+    static final int LONG = 5;
 
     /**
      * The type of CONSTANT_Double constant pool items.
      */
-    final static int DOUBLE = 6;
+    static final int DOUBLE = 6;
 
     /**
      * The type of CONSTANT_NameAndType constant pool items.
      */
-    final static int NAME_TYPE = 12;
+    static final int NAME_TYPE = 12;
 
     /**
      * The type of CONSTANT_Utf8 constant pool items.
      */
-    final static int UTF8 = 1;
+    static final int UTF8 = 1;
+
+    /**
+     * Normal type Item stored in the ClassWriter {@link ClassWriter#typeTable},
+     * instead of the constant pool, in order to avoid clashes with normal
+     * constant pool items in the ClassWriter constant pool's hash table.
+     */
+    static final int TYPE_NORMAL = 13;
+
+    /**
+     * Uninitialized type Item stored in the ClassWriter
+     * {@link ClassWriter#typeTable}, instead of the constant pool, in order to
+     * avoid clashes with normal constant pool items in the ClassWriter constant
+     * pool's hash table.
+     */
+    static final int TYPE_UNINIT = 14;
+
+    /**
+     * Merged type Item stored in the ClassWriter {@link ClassWriter#typeTable},
+     * instead of the constant pool, in order to avoid clashes with normal
+     * constant pool items in the ClassWriter constant pool's hash table.
+     */
+    static final int TYPE_MERGED = 15;
+
+    /**
+     * The class reader from which this class writer was constructed, if any.
+     */
+    ClassReader cr;
 
     /**
      * Minor and major version numbers of the class to be generated.
@@ -193,37 +245,55 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Index of the next item to be added in the constant pool.
      */
-    private short index;
+    int index;
 
     /**
      * The constant pool of this class.
      */
-    private ByteVector pool;
+    final ByteVector pool;
 
     /**
      * The constant pool's hash table data.
      */
-    private Item[] items;
+    Item[] items;
 
     /**
      * The threshold of the constant pool's hash table.
      */
-    private int threshold;
+    int threshold;
 
     /**
-     * A reusable key used to look for items in the hash {@link #items items}.
+     * A reusable key used to look for items in the {@link #items} hash table.
      */
-    Item key;
+    final Item key;
 
     /**
-     * A reusable key used to look for items in the hash {@link #items items}.
+     * A reusable key used to look for items in the {@link #items} hash table.
      */
-    Item key2;
+    final Item key2;
 
     /**
-     * A reusable key used to look for items in the hash {@link #items items}.
+     * A reusable key used to look for items in the {@link #items} hash table.
      */
-    Item key3;
+    final Item key3;
+
+    /**
+     * A type table used to temporarily store internal names that will not
+     * necessarily be stored in the constant pool. This type table is used by
+     * the control flow and data flow analysis algorithm used to compute stack
+     * map frames from scratch. This array associates to each index <tt>i</tt>
+     * the Item whose index is <tt>i</tt>. All Item objects stored in this
+     * array are also stored in the {@link #items} hash table. These two arrays
+     * allow to retrieve an Item from its index or, conversely, to get the index
+     * of an Item from its value. Each Item stores an internal name in its
+     * {@link Item#strVal1} field.
+     */
+    Item[] typeTable;
+
+    /**
+     * Number of elements in the {@link #typeTable} array.
+     */
+    private short typeCount;
 
     /**
      * The access flags of this class.
@@ -235,6 +305,11 @@ public class ClassWriter implements ClassVisitor {
      */
     private int name;
 
+    /**
+     * The internal name of this class.
+     */
+    String thisName;
+
     /**
      * The constant pool item that contains the signature of this class.
      */
@@ -342,12 +417,22 @@ public class ClassWriter implements ClassVisitor {
      * <tt>true</tt> if the maximum stack size and number of local variables
      * must be automatically computed.
      */
-    private boolean computeMaxs;
+    private final boolean computeMaxs;
 
     /**
-     * <tt>true</tt> to test that all attributes are known.
+     * <tt>true</tt> if the stack map frames must be recomputed from scratch.
      */
-    boolean checkAttributes;
+    private final boolean computeFrames;
+
+    /**
+     * <tt>true</tt> if the stack map tables of this class are invalid. The
+     * {@link MethodWriter#resizeInstructions} method cannot transform existing
+     * stack map tables, and so produces potentially invalid classes when it is
+     * executed. In this case the class is reread and rewritten with the
+     * {@link #COMPUTE_FRAMES} option (the resizeInstructions method can resize
+     * stack map tables when this option is used).
+     */
+    boolean invalidFrames;
 
     // ------------------------------------------------------------------------
     // Static initializer
@@ -362,7 +447,7 @@ public class ClassWriter implements ClassVisitor {
         String s = "AAAAAAAAAAAAAAAABCKLLDDDDDEEEEEEEEEEEEEEEEEEEEAAAAAAAADD"
                 + "DDDEEEEEEEEEEEEEEEEEEEEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
                 + "AAAAAAAAAAAAAAAAAMAAAAAAAAAAAAAAAAAAAAIIIIIIIIIIIIIIIIDNOAA"
-                + "AAAAGGGGGGGHAFBFAAFFAAQPIIJJIIIIIIIIIIIIIIIIII";
+                + "AAAAGGGGGGGHHFBFAAFFAAQPIIJJIIIIIIIIIIIIIIIIII";
         for (i = 0; i < b.length; ++i) {
             b[i] = (byte) (s.charAt(i) - 'A');
         }
@@ -402,7 +487,8 @@ public class ClassWriter implements ClassVisitor {
         // for (i = Constants.GETSTATIC; i <= Constants.INVOKESTATIC; ++i) {
         // b[i] = FIELDORMETH_INSN;
         // }
-        // b[Constants.INVOKEINTERFACE] = ITFMETH_INSN;
+        // b[Constants.INVOKEINTERFACE] = ITFDYNMETH_INSN;
+        // b[Constants.INVOKEDYNAMIC] = ITFDYNMETH_INSN;
         //
         // // LABEL(W)_INSN instructions
         // for (i = Constants.IFEQ; i <= Constants.JSR; ++i) {
@@ -440,41 +526,13 @@ public class ClassWriter implements ClassVisitor {
     // Constructor
     // ------------------------------------------------------------------------
 
-    /**
-     * Constructs a new {@link ClassWriter ClassWriter} object.
-     * 
-     * @param computeMaxs <tt>true</tt> if the maximum stack size and the
-     *        maximum number of local variables must be automatically computed.
-     *        If this flag is <tt>true</tt>, then the arguments of the
-     *        {@link MethodVisitor#visitMaxs visitMaxs} method of the
-     *        {@link MethodVisitor} returned by the
-     *        {@link #visitMethod visitMethod} method will be ignored, and
-     *        computed automatically from the signature and the bytecode of each
-     *        method.
-     */
-    public ClassWriter(final boolean computeMaxs) {
-        this(computeMaxs, false);
-    }
-
     /**
      * Constructs a new {@link ClassWriter} object.
-     * 
-     * @param computeMaxs <tt>true</tt> if the maximum stack size and the
-     *        maximum number of local variables must be automatically computed.
-     *        If this flag is <tt>true</tt>, then the arguments of the
-     *        {@link MethodVisitor#visitMaxs visitMaxs} method of the
-     *        {@link MethodVisitor} returned by the
-     *        {@link #visitMethod visitMethod} method will be ignored, and
-     *        computed automatically from the signature and the bytecode of each
-     *        method.
-     * @param skipUnknownAttributes <tt>true</tt> to silently ignore unknown
-     *        attributes, or <tt>false</tt> to throw an exception if an
-     *        unknown attribute is found.
+     *
+     * @param flags option flags that can be used to modify the default behavior
+     *        of this class. See {@link #COMPUTE_MAXS}, {@link #COMPUTE_FRAMES}.
      */
-    public ClassWriter(
-        final boolean computeMaxs,
-        final boolean skipUnknownAttributes)
-    {
+    public ClassWriter(final int flags) {
         index = 1;
         pool = new ByteVector();
         items = new Item[256];
@@ -482,8 +540,38 @@ public class ClassWriter implements ClassVisitor {
         key = new Item();
         key2 = new Item();
         key3 = new Item();
-        this.computeMaxs = computeMaxs;
-        this.checkAttributes = !skipUnknownAttributes;
+        this.computeMaxs = (flags & COMPUTE_MAXS) != 0;
+        this.computeFrames = (flags & COMPUTE_FRAMES) != 0;
+    }
+
+    /**
+     * Constructs a new {@link ClassWriter} object and enables optimizations for
+     * "mostly add" bytecode transformations. These optimizations are the
+     * following:
+     *
+     * <ul> <li>The constant pool from the original class is copied as is in
+     * the new class, which saves time. New constant pool entries will be added
+     * at the end if necessary, but unused constant pool entries <i>won't be
+     * removed</i>.</li> <li>Methods that are not transformed are copied as
+     * is in the new class, directly from the original class bytecode (i.e.
+     * without emitting visit events for all the method instructions), which
+     * saves a <i>lot</i> of time. Untransformed methods are detected by the
+     * fact that the {@link ClassReader} receives {@link MethodVisitor} objects
+     * that come from a {@link ClassWriter} (and not from a custom
+     * {@link ClassAdapter} or any other {@link ClassVisitor} instance).</li>
+     * </ul>
+     *
+     * @param classReader the {@link ClassReader} used to read the original
+     *        class. It will be used to copy the entire constant pool from the
+     *        original class and also to copy other fragments of original
+     *        bytecode where applicable.
+     * @param flags option flags that can be used to modify the default behavior
+     *        of this class. See {@link #COMPUTE_MAXS}, {@link #COMPUTE_FRAMES}.
+     */
+    public ClassWriter(final ClassReader classReader, final int flags) {
+        this(flags);
+        classReader.copyPool(this);
+        this.cr = classReader;
     }
 
     // ------------------------------------------------------------------------
@@ -501,7 +589,8 @@ public class ClassWriter implements ClassVisitor {
         this.version = version;
         this.access = access;
         this.name = newClass(name);
-        if (signature != null) {
+        thisName = name;
+        if (ClassReader.SIGNATURES && signature != null) {
             this.signature = newUTF8(signature);
         }
         this.superName = superName == null ? 0 : newClass(superName);
@@ -519,8 +608,7 @@ public class ClassWriter implements ClassVisitor {
             sourceFile = newUTF8(file);
         }
         if (debug != null) {
-            sourceDebug = new ByteVector();
-            sourceDebug.putUTF8(debug);
+            sourceDebug = new ByteVector().putUTF8(debug);
         }
     }
 
@@ -535,7 +623,13 @@ public class ClassWriter implements ClassVisitor {
         }
     }
 
-    public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
+    public AnnotationVisitor visitAnnotation(
+        final String desc,
+        final boolean visible)
+    {
+        if (!ClassReader.ANNOTATIONS) {
+            return null;
+        }
         ByteVector bv = new ByteVector();
         // write type, and reserve space for values count
         bv.putShort(newUTF8(desc)).putShort(0);
@@ -594,7 +688,8 @@ public class ClassWriter implements ClassVisitor {
                 desc,
                 signature,
                 exceptions,
-                computeMaxs);
+                computeMaxs,
+                computeFrames);
     }
 
     public void visitEnd() {
@@ -606,7 +701,7 @@ public class ClassWriter implements ClassVisitor {
 
     /**
      * Returns the bytecode of the class that was build with this class writer.
-     * 
+     *
      * @return the bytecode of the class that was build with this class writer.
      */
     public byte[] toByteArray() {
@@ -627,7 +722,7 @@ public class ClassWriter implements ClassVisitor {
             mb = mb.next;
         }
         int attributeCount = 0;
-        if (signature != 0) {
+        if (ClassReader.SIGNATURES && signature != 0) {
             ++attributeCount;
             size += 8;
             newUTF8("Signature");
@@ -639,7 +734,7 @@ public class ClassWriter implements ClassVisitor {
         }
         if (sourceDebug != null) {
             ++attributeCount;
-            size += sourceDebug.length;
+            size += sourceDebug.length + 4;
             newUTF8("SourceDebugExtension");
         }
         if (enclosingMethodOwner != 0) {
@@ -659,29 +754,17 @@ public class ClassWriter implements ClassVisitor {
             size += 6;
             newUTF8("Synthetic");
         }
-        if (version == Opcodes.V1_4) {
-            if ((access & Opcodes.ACC_ANNOTATION) != 0) {
-                ++attributeCount;
-                size += 6;
-                newUTF8("Annotation");
-            }
-            if ((access & Opcodes.ACC_ENUM) != 0) {
-                ++attributeCount;
-                size += 6;
-                newUTF8("Enum");
-            }
-        }
         if (innerClasses != null) {
             ++attributeCount;
             size += 8 + innerClasses.length;
             newUTF8("InnerClasses");
         }
-        if (anns != null) {
+        if (ClassReader.ANNOTATIONS && anns != null) {
             ++attributeCount;
             size += 8 + anns.getSize();
             newUTF8("RuntimeVisibleAnnotations");
         }
-        if (ianns != null) {
+        if (ClassReader.ANNOTATIONS && ianns != null) {
             ++attributeCount;
             size += 8 + ianns.getSize();
             newUTF8("RuntimeInvisibleAnnotations");
@@ -714,16 +797,16 @@ public class ClassWriter implements ClassVisitor {
             mb = mb.next;
         }
         out.putShort(attributeCount);
-        if (signature != 0) {
+        if (ClassReader.SIGNATURES && signature != 0) {
             out.putShort(newUTF8("Signature")).putInt(2).putShort(signature);
         }
         if (sourceFile != 0) {
             out.putShort(newUTF8("SourceFile")).putInt(2).putShort(sourceFile);
         }
         if (sourceDebug != null) {
-            int len = sourceDebug.length;
+            int len = sourceDebug.length - 2;
             out.putShort(newUTF8("SourceDebugExtension")).putInt(len);
-            out.putByteArray(sourceDebug.data, 0, len);
+            out.putByteArray(sourceDebug.data, 2, len);
         }
         if (enclosingMethodOwner != 0) {
             out.putShort(newUTF8("EnclosingMethod")).putInt(4);
@@ -737,30 +820,27 @@ public class ClassWriter implements ClassVisitor {
         {
             out.putShort(newUTF8("Synthetic")).putInt(0);
         }
-        if (version == Opcodes.V1_4) {
-            if ((access & Opcodes.ACC_ANNOTATION) != 0) {
-                out.putShort(newUTF8("Annotation")).putInt(0);
-            }
-            if ((access & Opcodes.ACC_ENUM) != 0) {
-                out.putShort(newUTF8("Enum")).putInt(0);
-            }
-        }
         if (innerClasses != null) {
             out.putShort(newUTF8("InnerClasses"));
             out.putInt(innerClasses.length + 2).putShort(innerClassesCount);
             out.putByteArray(innerClasses.data, 0, innerClasses.length);
         }
-        if (anns != null) {
+        if (ClassReader.ANNOTATIONS && anns != null) {
             out.putShort(newUTF8("RuntimeVisibleAnnotations"));
             anns.put(out);
         }
-        if (ianns != null) {
+        if (ClassReader.ANNOTATIONS && ianns != null) {
             out.putShort(newUTF8("RuntimeInvisibleAnnotations"));
             ianns.put(out);
         }
         if (attrs != null) {
             attrs.put(this, null, 0, -1, -1, out);
         }
+        if (invalidFrames) {
+            ClassWriter cw = new ClassWriter(COMPUTE_FRAMES);
+            new ClassReader(out.data).accept(cw, ClassReader.SKIP_FRAMES);
+            return cw.toByteArray();
+        }
         return out.data;
     }
 
@@ -771,7 +851,7 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Adds a number or string constant to the constant pool of the class being
      * build. Does nothing if the constant pool already contains a similar item.
-     * 
+     *
      * @param cst the value of the constant to be added to the constant pool.
      *        This parameter must be an {@link Integer}, a {@link Float}, a
      *        {@link Long}, a {@link Double}, a {@link String} or a
@@ -820,7 +900,7 @@ public class ClassWriter implements ClassVisitor {
      * build. Does nothing if the constant pool already contains a similar item.
      * <i>This method is intended for {@link Attribute} sub classes, and is
      * normally not needed by class generators or adapters.</i>
-     * 
+     *
      * @param cst the value of the constant to be added to the constant pool.
      *        This parameter must be an {@link Integer}, a {@link Float}, a
      *        {@link Long}, a {@link Double} or a {@link String}.
@@ -836,12 +916,12 @@ public class ClassWriter implements ClassVisitor {
      * nothing if the constant pool already contains a similar item. <i>This
      * method is intended for {@link Attribute} sub classes, and is normally not
      * needed by class generators or adapters.</i>
-     * 
+     *
      * @param value the String value.
      * @return the index of a new or already existing UTF8 item.
      */
     public int newUTF8(final String value) {
-        key.set('s', value, null, null);
+        key.set(UTF8, value, null, null);
         Item result = get(key);
         if (result == null) {
             pool.putByte(UTF8).putUTF8(value);
@@ -856,25 +936,12 @@ public class ClassWriter implements ClassVisitor {
      * Does nothing if the constant pool already contains a similar item.
      * <i>This method is intended for {@link Attribute} sub classes, and is
      * normally not needed by class generators or adapters.</i>
-     * 
-     * @param value the internal name of the class.
-     * @return the index of a new or already existing class reference item.
-     */
-    public int newClass(final String value) {
-        return newClassItem(value).index;
-    }
-
-    /**
-     * Adds a class reference to the constant pool of the class being build.
-     * Does nothing if the constant pool already contains a similar item.
-     * <i>This method is intended for {@link Attribute} sub classes, and is
-     * normally not needed by class generators or adapters.</i>
-     * 
+     *
      * @param value the internal name of the class.
      * @return a new or already existing class reference item.
      */
-    private Item newClassItem(final String value) {
-        key2.set('C', value, null, null);
+    Item newClassItem(final String value) {
+        key2.set(CLASS, value, null, null);
         Item result = get(key2);
         if (result == null) {
             pool.put12(CLASS, newUTF8(value));
@@ -884,12 +951,46 @@ public class ClassWriter implements ClassVisitor {
         return result;
     }
 
+    /**
+     * Adds a class reference to the constant pool of the class being build.
+     * Does nothing if the constant pool already contains a similar item.
+     * <i>This method is intended for {@link Attribute} sub classes, and is
+     * normally not needed by class generators or adapters.</i>
+     *
+     * @param value the internal name of the class.
+     * @return the index of a new or already existing class reference item.
+     */
+    public int newClass(final String value) {
+        return newClassItem(value).index;
+    }
+
+    /**
+     * Adds a field reference to the constant pool of the class being build.
+     * Does nothing if the constant pool already contains a similar item.
+     *
+     * @param owner the internal name of the field's owner class.
+     * @param name the field's name.
+     * @param desc the field's descriptor.
+     * @return a new or already existing field reference item.
+     */
+    Item newFieldItem(final String owner, final String name, final String desc)
+    {
+        key3.set(FIELD, owner, name, desc);
+        Item result = get(key3);
+        if (result == null) {
+            put122(FIELD, newClass(owner), newNameType(name, desc));
+            result = new Item(index++, key3);
+            put(result);
+        }
+        return result;
+    }
+
     /**
      * Adds a field reference to the constant pool of the class being build.
      * Does nothing if the constant pool already contains a similar item.
      * <i>This method is intended for {@link Attribute} sub classes, and is
      * normally not needed by class generators or adapters.</i>
-     * 
+     *
      * @param owner the internal name of the field's owner class.
      * @param name the field's name.
      * @param desc the field's descriptor.
@@ -897,20 +998,13 @@ public class ClassWriter implements ClassVisitor {
      */
     public int newField(final String owner, final String name, final String desc)
     {
-        key3.set('G', owner, name, desc);
-        Item result = get(key3);
-        if (result == null) {
-            put122(FIELD, newClass(owner), newNameType(name, desc));
-            result = new Item(index++, key3);
-            put(result);
-        }
-        return result.index;
+        return newFieldItem(owner, name, desc).index;
     }
 
     /**
      * Adds a method reference to the constant pool of the class being build.
      * Does nothing if the constant pool already contains a similar item.
-     * 
+     *
      * @param owner the internal name of the method's owner class.
      * @param name the method's name.
      * @param desc the method's descriptor.
@@ -923,10 +1017,11 @@ public class ClassWriter implements ClassVisitor {
         final String desc,
         final boolean itf)
     {
-        key3.set(itf ? 'N' : 'M', owner, name, desc);
+        int type = itf ? IMETH : METH;
+        key3.set(type, owner, name, desc);
         Item result = get(key3);
         if (result == null) {
-            put122(itf ? IMETH : METH, newClass(owner), newNameType(name, desc));
+            put122(type, newClass(owner), newNameType(name, desc));
             result = new Item(index++, key3);
             put(result);
         }
@@ -938,7 +1033,7 @@ public class ClassWriter implements ClassVisitor {
      * Does nothing if the constant pool already contains a similar item.
      * <i>This method is intended for {@link Attribute} sub classes, and is
      * normally not needed by class generators or adapters.</i>
-     * 
+     *
      * @param owner the internal name of the method's owner class.
      * @param name the method's name.
      * @param desc the method's descriptor.
@@ -957,7 +1052,7 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Adds an integer to the constant pool of the class being build. Does
      * nothing if the constant pool already contains a similar item.
-     * 
+     *
      * @param value the int value.
      * @return a new or already existing int item.
      */
@@ -975,7 +1070,7 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Adds a float to the constant pool of the class being build. Does nothing
      * if the constant pool already contains a similar item.
-     * 
+     *
      * @param value the float value.
      * @return a new or already existing float item.
      */
@@ -983,7 +1078,7 @@ public class ClassWriter implements ClassVisitor {
         key.set(value);
         Item result = get(key);
         if (result == null) {
-            pool.putByte(FLOAT).putInt(Float.floatToIntBits(value));
+            pool.putByte(FLOAT).putInt(key.intVal);
             result = new Item(index++, key);
             put(result);
         }
@@ -993,7 +1088,7 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Adds a long to the constant pool of the class being build. Does nothing
      * if the constant pool already contains a similar item.
-     * 
+     *
      * @param value the long value.
      * @return a new or already existing long item.
      */
@@ -1012,7 +1107,7 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Adds a double to the constant pool of the class being build. Does nothing
      * if the constant pool already contains a similar item.
-     * 
+     *
      * @param value the double value.
      * @return a new or already existing double item.
      */
@@ -1020,7 +1115,7 @@ public class ClassWriter implements ClassVisitor {
         key.set(value);
         Item result = get(key);
         if (result == null) {
-            pool.putByte(DOUBLE).putLong(Double.doubleToLongBits(value));
+            pool.putByte(DOUBLE).putLong(key.longVal);
             result = new Item(index, key);
             put(result);
             index += 2;
@@ -1031,12 +1126,12 @@ public class ClassWriter implements ClassVisitor {
     /**
      * Adds a string to the constant pool of the class being build. Does nothing
      * if the constant pool already contains a similar item.
-     * 
+     *
      * @param value the String value.
      * @return a new or already existing string item.
      */
     private Item newString(final String value) {
-        key2.set('S', value, null, null);
+        key2.set(STR, value, null, null);
         Item result = get(key2);
         if (result == null) {
             pool.put12(STR, newUTF8(value));
@@ -1051,52 +1146,187 @@ public class ClassWriter implements ClassVisitor {
      * nothing if the constant pool already contains a similar item. <i>This
      * method is intended for {@link Attribute} sub classes, and is normally not
      * needed by class generators or adapters.</i>
-     * 
+     *
      * @param name a name.
      * @param desc a type descriptor.
      * @return the index of a new or already existing name and type item.
      */
     public int newNameType(final String name, final String desc) {
-        key2.set('T', name, desc, null);
+        return newNameTypeItem(name, desc).index;
+    }
+    
+    /**
+     * Adds a name and type to the constant pool of the class being build. Does
+     * nothing if the constant pool already contains a similar item.
+     *
+     * @param name a name.
+     * @param desc a type descriptor.
+     * @return a new or already existing name and type item.
+     */
+    Item newNameTypeItem(final String name, final String desc) {
+        key2.set(NAME_TYPE, name, desc, null);
         Item result = get(key2);
         if (result == null) {
             put122(NAME_TYPE, newUTF8(name), newUTF8(desc));
             result = new Item(index++, key2);
             put(result);
         }
+        return result;
+    }
+
+    /**
+     * Adds the given internal name to {@link #typeTable} and returns its index.
+     * Does nothing if the type table already contains this internal name.
+     *
+     * @param type the internal name to be added to the type table.
+     * @return the index of this internal name in the type table.
+     */
+    int addType(final String type) {
+        key.set(TYPE_NORMAL, type, null, null);
+        Item result = get(key);
+        if (result == null) {
+            result = addType(key);
+        }
         return result.index;
     }
 
+    /**
+     * Adds the given "uninitialized" type to {@link #typeTable} and returns its
+     * index. This method is used for UNINITIALIZED types, made of an internal
+     * name and a bytecode offset.
+     *
+     * @param type the internal name to be added to the type table.
+     * @param offset the bytecode offset of the NEW instruction that created
+     *        this UNINITIALIZED type value.
+     * @return the index of this internal name in the type table.
+     */
+    int addUninitializedType(final String type, final int offset) {
+        key.type = TYPE_UNINIT;
+        key.intVal = offset;
+        key.strVal1 = type;
+        key.hashCode = 0x7FFFFFFF & (TYPE_UNINIT + type.hashCode() + offset);
+        Item result = get(key);
+        if (result == null) {
+            result = addType(key);
+        }
+        return result.index;
+    }
+
+    /**
+     * Adds the given Item to {@link #typeTable}.
+     *
+     * @param item the value to be added to the type table.
+     * @return the added Item, which a new Item instance with the same value as
+     *         the given Item.
+     */
+    private Item addType(final Item item) {
+        ++typeCount;
+        Item result = new Item(typeCount, key);
+        put(result);
+        if (typeTable == null) {
+            typeTable = new Item[16];
+        }
+        if (typeCount == typeTable.length) {
+            Item[] newTable = new Item[2 * typeTable.length];
+            System.arraycopy(typeTable, 0, newTable, 0, typeTable.length);
+            typeTable = newTable;
+        }
+        typeTable[typeCount] = result;
+        return result;
+    }
+
+    /**
+     * Returns the index of the common super type of the two given types. This
+     * method calls {@link #getCommonSuperClass} and caches the result in the
+     * {@link #items} hash table to speedup future calls with the same
+     * parameters.
+     *
+     * @param type1 index of an internal name in {@link #typeTable}.
+     * @param type2 index of an internal name in {@link #typeTable}.
+     * @return the index of the common super type of the two given types.
+     */
+    int getMergedType(final int type1, final int type2) {
+        key2.type = TYPE_MERGED;
+        key2.longVal = type1 | (((long) type2) << 32);
+        key2.hashCode = 0x7FFFFFFF & (TYPE_MERGED + type1 + type2);
+        Item result = get(key2);
+        if (result == null) {
+            String t = typeTable[type1].strVal1;
+            String u = typeTable[type2].strVal1;
+            key2.intVal = addType(getCommonSuperClass(t, u));
+            result = new Item((short) 0, key2);
+            put(result);
+        }
+        return result.intVal;
+    }
+
+    /**
+     * Returns the common super type of the two given types. The default
+     * implementation of this method <i>loads<i> the two given classes and uses
+     * the java.lang.Class methods to find the common super class. It can be
+     * overridden to compute this common super type in other ways, in particular
+     * without actually loading any class, or to take into account the class
+     * that is currently being generated by this ClassWriter, which can of
+     * course not be loaded since it is under construction.
+     *
+     * @param type1 the internal name of a class.
+     * @param type2 the internal name of another class.
+     * @return the internal name of the common super class of the two given
+     *         classes.
+     */
+    protected String getCommonSuperClass(final String type1, final String type2)
+    {
+        Class c, d;
+        try {
+            c = Class.forName(type1.replace('/', '.'));
+            d = Class.forName(type2.replace('/', '.'));
+        } catch (Exception e) {
+            throw new RuntimeException(e.toString());
+        }
+        if (c.isAssignableFrom(d)) {
+            return type1;
+        }
+        if (d.isAssignableFrom(c)) {
+            return type2;
+        }
+        if (c.isInterface() || d.isInterface()) {
+            return "java/lang/Object";
+        } else {
+            do {
+                c = c.getSuperclass();
+            } while (!c.isAssignableFrom(d));
+            return c.getName().replace('.', '/');
+        }
+    }
+
     /**
      * Returns the constant pool's hash table item which is equal to the given
      * item.
-     * 
+     *
      * @param key a constant pool item.
      * @return the constant pool's hash table item which is equal to the given
      *         item, or <tt>null</tt> if there is no such item.
      */
     private Item get(final Item key) {
-        int h = key.hashCode;
-        Item i = items[h % items.length];
-        while (i != null) {
-            if (i.hashCode == h && key.isEqualTo(i)) {
-                return i;
-            }
+        Item i = items[key.hashCode % items.length];
+        while (i != null && (i.type != key.type || !key.isEqualTo(i))) {
             i = i.next;
         }
-        return null;
+        return i;
     }
 
     /**
      * Puts the given item in the constant pool's hash table. The hash table
      * <i>must</i> not already contains this item.
-     * 
+     *
      * @param i the item to be added to the constant pool's hash table.
      */
     private void put(final Item i) {
         if (index > threshold) {
-            Item[] newItems = new Item[items.length * 2 + 1];
-            for (int l = items.length - 1; l >= 0; --l) {
+            int ll = items.length;
+            int nl = ll * 2 + 1;
+            Item[] newItems = new Item[nl];
+            for (int l = ll - 1; l >= 0; --l) {
                 Item j = items[l];
                 while (j != null) {
                     int index = j.hashCode % newItems.length;
@@ -1107,7 +1337,7 @@ public class ClassWriter implements ClassVisitor {
                 }
             }
             items = newItems;
-            threshold = (int) (items.length * 0.75);
+            threshold = (int) (nl * 0.75);
         }
         int index = i.hashCode % items.length;
         i.next = items[index];
@@ -1116,7 +1346,7 @@ public class ClassWriter implements ClassVisitor {
 
     /**
      * Puts one byte and two shorts into the constant pool.
-     * 
+     *
      * @param b a byte.
      * @param s1 a short.
      * @param s2 another short.
