@@ -1,6 +1,6 @@
 /***
  * ASM: a very small and fast Java bytecode manipulation framework
- * Copyright (c) 2000-2005 INRIA, France Telecom
+ * Copyright (c) 2000-2007 INRIA, France Telecom
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,10 +30,10 @@
 package org.objectweb.asm.commons;
 
 import java.io.ByteArrayOutputStream;
+import java.io.DataOutput;
 import java.io.DataOutputStream;
 import java.io.IOException;
 import java.security.MessageDigest;
-import java.security.NoSuchAlgorithmException;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
@@ -46,7 +46,16 @@ import org.objectweb.asm.Opcodes;
 
 /**
  * A {@link ClassAdapter} that adds a serial version unique identifier to a
- * class if missing. The SVUID algorithm can be found <a href=
+ * class if missing. Here is typical usage of this class:
+ * 
+ * <pre>
+ *   ClassWriter cw = new ClassWriter(...);
+ *   ClassVisitor sv = new SerialVersionUIDAdder(cw);
+ *   ClassVisitor ca = new MyClassAdapter(sv);
+ *   new ClassReader(orginalClass).accept(ca, false);
+ * </pre>
+ * 
+ * The SVUID algorithm can be found <a href=
  * "http://java.sun.com/j2se/1.4.2/docs/guide/serialization/spec/class.html"
  * >http://java.sun.com/j2se/1.4.2/docs/guide/serialization/spec/class.html</a>:
  * 
@@ -168,8 +177,8 @@ public class SerialVersionUIDAdder extends ClassAdapter {
     // ------------------------------------------------------------------------
 
     /*
-     * Visit class header and get class name, access , and intefraces
-     * informatoin (step 1,2, and 3) for SVUID computation.
+     * Visit class header and get class name, access , and interfaces
+     * information (step 1,2, and 3) for SVUID computation.
      */
     public void visit(
         final int version,
@@ -192,7 +201,7 @@ public class SerialVersionUIDAdder extends ClassAdapter {
 
     /*
      * Visit the methods and get constructor and method information (step 5 and
-     * 7). Also determince if there is a class initializer (step 6).
+     * 7). Also determine if there is a class initializer (step 6).
      */
     public MethodVisitor visitMethod(
         final int access,
@@ -202,7 +211,7 @@ public class SerialVersionUIDAdder extends ClassAdapter {
         final String[] exceptions)
     {
         if (computeSVUID) {
-            if (name.equals("<clinit>")) {
+            if ("<clinit>".equals(name)) {
                 hasStaticInitializer = true;
             }
             /*
@@ -220,9 +229,9 @@ public class SerialVersionUIDAdder extends ClassAdapter {
 
             // all non private methods
             if ((access & Opcodes.ACC_PRIVATE) == 0) {
-                if (name.equals("<init>")) {
+                if ("<init>".equals(name)) {
                     svuidConstructors.add(new Item(name, mods, desc));
-                } else if (!name.equals("<clinit>")) {
+                } else if (!"<clinit>".equals(name)) {
                     svuidMethods.add(new Item(name, mods, desc));
                 }
             }
@@ -232,7 +241,7 @@ public class SerialVersionUIDAdder extends ClassAdapter {
     }
 
     /*
-     * Gets class field information for step 4 of the alogrithm. Also determines
+     * Gets class field information for step 4 of the algorithm. Also determines
      * if the class already has a SVUID.
      */
     public FieldVisitor visitField(
@@ -243,7 +252,7 @@ public class SerialVersionUIDAdder extends ClassAdapter {
         final Object value)
     {
         if (computeSVUID) {
-            if (name.equals("serialVersionUID")) {
+            if ("serialVersionUID".equals(name)) {
                 // since the class already has SVUID, we won't be computing it.
                 computeSVUID = false;
                 hasSVUID = true;
@@ -259,8 +268,8 @@ public class SerialVersionUIDAdder extends ClassAdapter {
                             | Opcodes.ACC_PROTECTED | Opcodes.ACC_STATIC
                             | Opcodes.ACC_FINAL | Opcodes.ACC_VOLATILE | Opcodes.ACC_TRANSIENT);
 
-            if (((access & Opcodes.ACC_PRIVATE) == 0)
-                    || ((access & (Opcodes.ACC_STATIC | Opcodes.ACC_TRANSIENT)) == 0))
+            if ((access & Opcodes.ACC_PRIVATE) == 0
+                    || (access & (Opcodes.ACC_STATIC | Opcodes.ACC_TRANSIENT)) == 0)
             {
                 svuidFields.add(new Item(name, mods, desc));
             }
@@ -269,6 +278,28 @@ public class SerialVersionUIDAdder extends ClassAdapter {
         return super.visitField(access, name, desc, signature, value);
     }
 
+    /**
+     * Handle a bizarre special case. Nested classes (static classes declared
+     * inside another class) that are protected have their access bit set to
+     * public in their class files to deal with some odd reflection situation.
+     * Our SVUID computation must do as the JVM does and ignore access bits in
+     * the class file in favor of the access bits InnerClass attribute.
+     */
+    public void visitInnerClass(final String aname, final String outerName, final String innerName, final int attr_access) {
+        if (outerName != null && innerName != null && name != null) {
+            int len = name.length();
+            int ilen = innerName.length();
+            int olen = outerName.length();
+            if (len == olen + 1 + ilen 
+                    && this.name.startsWith(outerName) && this.name.endsWith(innerName) 
+                    && this.name.charAt(olen) == '$') 
+            {
+                this.access = attr_access; 
+            }
+        }
+        super.visitInnerClass(aname, outerName, innerName, attr_access);
+    }
+    
     /*
      * Add the SVUID if class doesn't have one
      */
@@ -300,15 +331,9 @@ public class SerialVersionUIDAdder extends ClassAdapter {
      * <code>isHasSVUID</code> to determine if the class already had an SVUID.
      * 
      * @return Returns the serial version UID
-     * @throws IOException
-     * @throws NoSuchAlgorithmException
      */
-    protected long computeSVUID() throws IOException, NoSuchAlgorithmException {
-        if (hasSVUID) {
-            return 0;
-        }
-
-        ByteArrayOutputStream bos = null;
+    protected long computeSVUID() throws IOException {
+        ByteArrayOutputStream bos;
         DataOutputStream dos = null;
         long svuid = 0;
 
@@ -345,7 +370,7 @@ public class SerialVersionUIDAdder extends ClassAdapter {
              * field written as a 32-bit integer. 3. The descriptor of the field
              * in UTF encoding
              * 
-             * Note that field signatutes are not dot separated. Method and
+             * Note that field signatures are not dot separated. Method and
              * constructor signatures are dot separated. Go figure...
              */
             writeItems(svuidFields, dos, false);
@@ -386,7 +411,7 @@ public class SerialVersionUIDAdder extends ClassAdapter {
              * produced by DataOutputStream and produces five 32-bit values
              * sha[0..4].
              */
-            MessageDigest md = MessageDigest.getInstance("SHA");
+            byte[] hashBytes = computeSHAdigest(bos.toByteArray());
 
             /*
              * 9. The hash value is assembled from the first and second 32-bit
@@ -401,7 +426,6 @@ public class SerialVersionUIDAdder extends ClassAdapter {
              * 40 | ((sha[1] >>> 8) & 0xFF) << 48 | ((sha[1] >>> 0) & 0xFF) <<
              * 56;
              */
-            byte[] hashBytes = md.digest(bos.toByteArray());
             for (int i = Math.min(hashBytes.length, 8) - 1; i >= 0; i--) {
                 svuid = (svuid << 8) | (hashBytes[i] & 0xFF);
             }
@@ -415,6 +439,20 @@ public class SerialVersionUIDAdder extends ClassAdapter {
         return svuid;
     }
 
+    /**
+     * Returns the SHA-1 message digest of the given value.
+     * 
+     * @param value the value whose SHA message digest must be computed.
+     * @return the SHA-1 message digest of the given value.
+     */
+    protected byte[] computeSHAdigest(final byte[] value) {
+        try {
+            return MessageDigest.getInstance("SHA").digest(value);
+        } catch (Exception e) {
+            throw new UnsupportedOperationException(e.toString());
+        }
+    }
+
     /**
      * Sorts the items in the collection and writes it to the data output stream
      * 
@@ -423,13 +461,13 @@ public class SerialVersionUIDAdder extends ClassAdapter {
      * @param dotted a <code>boolean</code> value
      * @exception IOException if an error occurs
      */
-    private void writeItems(
+    private static void writeItems(
         final Collection itemCollection,
-        final DataOutputStream dos,
+        final DataOutput dos,
         final boolean dotted) throws IOException
     {
         int size = itemCollection.size();
-        Item items[] = (Item[]) itemCollection.toArray(new Item[size]);
+        Item[] items = (Item[]) itemCollection.toArray(new Item[size]);
         Arrays.sort(items);
         for (int i = 0; i < size; i++) {
             dos.writeUTF(items[i].name);
@@ -446,11 +484,11 @@ public class SerialVersionUIDAdder extends ClassAdapter {
 
     static class Item implements Comparable {
 
-        String name;
+        final String name;
 
-        int access;
+        final int access;
 
-        String desc;
+        final String desc;
 
         Item(final String name, final int access, final String desc) {
             this.name = name;
